﻿using System;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace blue_dish_te
{
	internal class LahevObject : MapObject
	{
		private ExtendedMaterial[] eMaterials = null;
		private Material[] materials;
		private Texture[] textures;
		private Mesh mesh;
		private float rotateTime = 0;

		public const float cRotateTimer = 3f;
		public const float cRegenerationWaiting = 5f;
		public const float cRegeneration = 2f;

		private bool isPlayerOn = false;
		private bool picked = false;
		private float pickedWaiting = 0;
		private float regenerationTimer = 0;

		// Konstruktor objektu
		public LahevObject(Device dev, DatLoader dat, char type) : base(dev, dat, type)
		{
			try
			{
				mesh = Mesh.FromStream(dat.GetItem("meshes/bottle.x"), MeshFlags.Managed, dev, out eMaterials);
			}
			catch(Exception e)
			{
				MessageBox.Show("Nepovedlo se nacist objekt Lahve.");
				Console.WriteLine(e);
				return;
			}
			//Create an array of textures (Our model/mesh may have different textures assigned to it)
			textures = new Texture[eMaterials.Length];
			//Create an array of materials (Our model/mesh may have different materials assigned to it)
			materials = new Material[eMaterials.Length];
			for (int i = 0; i < materials.Length; i++)
			{
				//Retrieves or sets the Material structure that describes the material properties.
				materials[i] = eMaterials[i].Material3D;
				materials[i].Ambient = materials[i].Diffuse;
				//Retrieves or sets a string that specifies the file name of the texture.
				//Load the texture from the filename retrieved.
				try
				{
					if (eMaterials[i].TextureFilename != null)
						textures[i] = TextureLoader.FromStream(dev, dat.GetItem(eMaterials[i].TextureFilename));
				}
				catch(Exception e)
				{
					MessageBox.Show("Nepovedlo se nacist textury pro Lahev.");
					Console.WriteLine(e);
					return;
				}
			}
		}

		// Rendrovaci funkce
		public override void Render()
		{
			// je-li sebrano, pocitej timer do obnoveni objektu
			if (picked)
			{
				if (pickedWaiting > cRegenerationWaiting)
				{
					// objekt se ma zvetsovat
					regenerationTimer += renderTime;
					if (regenerationTimer > cRegeneration)
					{
						picked = false;
						pickedWaiting = 0;
						regenerationTimer = 0;
					}
				}
				else
				{
					pickedWaiting += renderTime;
				}
			}
			else
			{
				// pokud hrac neodesel z kosticky, sebere objekt znovu
				if (isPlayerOn)
				{
					pickedWaiting = 0;
					regenerationTimer = 0;
					picked = true;
					svet.beerAdd();
				}
			}

			rotateTime += renderTime;
			// pokud timer rotacniho casu dosel do maxima, vynulujeme ho
			if (rotateTime > cRotateTimer)
				rotateTime = 0;

			// pokud je objekt sebran a jeho timer zvetsovani je nulovy, tak ho nekreslime
			if (picked && regenerationTimer == 0)
				return;

			// provedeme prislusne translace a rotace, aby to vypadalo dobre
			float pos = rotateTime / cRotateTimer;
			float ypos = 1;

			if (picked)
			{
				ypos = regenerationTimer / cRegeneration;
			}

			Matrix m = device.Transform.World;
			device.Transform.World = Matrix.Translation(0, -.25f, 0) * Matrix.Scaling(ypos, ypos, ypos) * Matrix.RotationZ((float) -Math.PI / 4) * Matrix.RotationY(pos * 2F * (float) Math.PI) * Matrix.Translation(1, 0.5f + ypos * (0.75f + (float) Math.Sin(pos * 2 * (float) Math.PI) * 0.25f), 1) * m;
			Material oldMaterial;

			for (int i = 0; i < materials.Length; i++)
			{
				oldMaterial = device.Material;
				device.Material = materials[i];
				if (!noTexture) device.SetTexture(0, textures[i]);
				/* Draws a subset of a mesh:
                * An attribute table is used to identify areas of
                the mesh that need to be drawn with different textures,
                render states, and materials. The application can use
                the attribute table to hide portions of a mesh by
                leaving out a given attribute identifier (attributeID)
                when drawing the frame. Attribute tables
                can also be used for application defined purposes.*/
				mesh.DrawSubset(i);
				device.Material = oldMaterial;
			}
		}

		// hrac prisel na kosticku
		public override bool PlayerOver()
		{
			isPlayerOn = true;
			if (!picked) picked = true;
			svet.beerAdd();
			return false;
		}

		//hrac odesel z kosticky
		public override bool PlayerOut()
		{
			isPlayerOn = false;
			return false;
		}
	}
}