using System;
using System.ComponentModel;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using Microsoft.DirectX.DirectInput;
using WinFont = System.Drawing.Font;
using D3DFont = Microsoft.DirectX.Direct3D.Font;
using Device=Microsoft.DirectX.DirectInput.Device;
using DeviceType=Microsoft.DirectX.Direct3D.DeviceType;
using DInput = Microsoft.DirectX.DirectInput;
using Manager=Microsoft.DirectX.Direct3D.Manager;

namespace blue_dish_te
{
	/// <summary>
	/// Formular, ktery vytvori D3D device a ve smycce se prekresluje
	/// </summary>
	public class RenderForm : Form
	{
		[DllImport("user32.dll")]
		public static extern int SendNotifyMessage(IntPtr hWnd, int Msg, IntPtr wParam, IntPtr lParam);

		// promenne
		public DatLoader dat;
		private static Microsoft.DirectX.Direct3D.Device device = null;

		private Svet svet = null;
		private Kolize kolize = null;

		private Device mouse = null;
		private Device keyboard = null;
		private int mouseX = 0;
		private int mouseY = 10;
		private bool drzKrysu = false;

		private long tick = 0;
		private long LastTick = 0;
		private int loopsCount = 0;
		private int fps = 0;
		private float dobaVykreslovani = 0.5F;

		private bool invertMouse = false;

		private WinFont winFont;
		private D3DFont d3dFont;

		private Material material;

		private const float camFov = (float) Math.PI / 4;

		private const float camNear = 0.1f;
		private const float camFar = 100f;

		private bool bRenderWireFrame = false;
		private bool bRenderNoLights = false;
		private bool bRenderNoFog = false;
		private bool bRenderNoTextures = false;

		/// <summary>
		/// uhly pohledu na scenu
		/// </summary>
		private float azimuth;

		private float zenith;

		/// <summary>
		/// Konstruktor nastavi velikost a vlastnosti okna
		/// </summary>
		public RenderForm()
		{
			SetStyle(ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint, true);
			ComponentResourceManager resources = new ComponentResourceManager(typeof(RenderForm));

			ClientSize = new Size(640, 480); //velikost okna
			Text = "Blue-dish-te ver 1.0";
			Icon = ((Icon) (resources.GetObject("$this.Icon")));
			//this.FormBorderStyle = FormBorderStyle.FixedSingle; //styl ramecku okna
			//MaximizeBox = false; //minim. a maxim. tlacitko schovane
			//MinimizeBox = false;
			BringToFront();
			BackColor = Color.Black;
		}

		/**
         * Zjisteni zda grafika podporuje Hardwarovou akceleraci
         * Caps caps = Manager.GetDeviceCaps(0,DeviceType.Hardware );
         * if (caps.DeviceCaps.SupportsHardwareTransformAndLight) {
         *     ano;
         * else
         *     ne;
         **/

		/// <summary>
		/// Metoda se pokusi inicializovat D3D. V pripade neuspechu vypise na konzoli chybu
		/// </summary>
		/// <returns>V pripade neuspechu vraci FALSE jinak TRUE</returns>
		private bool D3DInit()
		{
			try
			{
				//otestujeme schopnosti grafiky
				Caps DevCaps = Manager.GetDeviceCaps(0, DeviceType.Hardware);
				DeviceType DevType = DevCaps.DeviceType;
				CreateFlags DevFlags = CreateFlags.SoftwareVertexProcessing;
				if ((DevCaps.VertexShaderVersion >= new Version(2, 0)) && (DevCaps.PixelShaderVersion >= new Version(2, 0)))
				{
					DevType = DeviceType.Hardware;
					if (DevCaps.DeviceCaps.SupportsHardwareTransformAndLight)
					{
						DevFlags = CreateFlags.HardwareVertexProcessing;
						Console.WriteLine("true");
					}
				}
				//DevFlags |= CreateFlags.MultiThreaded;

				//pripravime parametry
				PresentParameters param = new PresentParameters();
				param.EnableAutoDepthStencil = true;
				param.AutoDepthStencilFormat = DepthFormat.D16;
				param.SwapEffect = SwapEffect.Discard;
				param.BackBufferCount = 1;
				param.BackBufferWidth = Size.Width;
				param.BackBufferHeight = Size.Height;
				param.PresentationInterval = PresentInterval.Immediate;
				param.BackBufferFormat = Format.X8R8G8B8; //Manager.Adapters[0].CurrentDisplayMode.Format;
				param.Windowed = true;

				//a vytvorime device
				device = new Microsoft.DirectX.Direct3D.Device(0, DevType, this, DevFlags, param);

				//osetreni resetovani Device
				device.DeviceReset += new EventHandler(DeviceResetEvent);

				device.RenderState.CullMode = Cull.CounterClockwise; //orezavani odvracenych sten
				device.RenderState.ZBufferEnable = true;
				device.RenderState.Lighting = true; //zapnuti svetel
				device.RenderState.Ambient = Color.FromArgb(0x20, 0x20, 0x20);
				device.RenderState.SpecularEnable = true; //zapnuti spekularnich odlesku

				//nastaveni projekcni matice
				device.Transform.Projection = Matrix.PerspectiveFovLH(camFov, (float) ClientSize.Width / ClientSize.Height, camNear, camFar);

				//nastaveni nejlepsiho filtrovani textury
				setBestTextureFilters();
				
				device.RenderState.FogColor = Color.Black;
				device.RenderState.FogStart = camNear;
				device.RenderState.FogEnd = camFar;
				device.RenderState.FogDensity = 0.05f;
				device.RenderState.FogVertexMode = FogMode.Exp;
				device.RenderState.FogEnable = true;

				//vytvorime svetla
				CreateLights();

				//inicializace fontu pro fps
				winFont = new WinFont(FontFamily.GenericSansSerif, 12);
				d3dFont = new D3DFont(device, winFont);

				//nacteni datoveho souboru
				dat = DatLoader.LoadDataFile("blue-dish-te.dat");

				//nacteme svet
				svet = new Svet(device, dat);
				svet.LoadMap("semestralka.map");
				svet.SetRenderForm(this);

				//inicializace detekce kolizi
				kolize = new Kolize(svet);

				//inicializace materialu
				material = new Material();
				material.Diffuse = Color.White; //nastaveni difuzni barvy materialu
				//material.Specular = Color.White; //nastaveni spekularni barvy materialu
				material.Ambient = Color.White; //nastaveni ambientni barvy materialu
				material.Emissive = Color.FromArgb(10, 10, 10);
				//material.SpecularSharpness = .2f;

				Cursor.Position = new Point(Left + Width / 2, Top + Height / 2);
				Cursor.Hide();
				drzKrysu = true;

				return true;
			}
			catch(Exception e) //pokud doslo k vyjimce vrat FALSE
			{
				MessageBox.Show(e.ToString());
				Console.WriteLine(e);
				return false;
			}
		}

		private void DeviceResetEvent(object caller, EventArgs args)
		{
			device.RenderState.CullMode = Cull.CounterClockwise; //orezavani odvracenych sten
			device.RenderState.ZBufferEnable = true;
			device.RenderState.Lighting = true; //zapnuti svetel
			device.RenderState.Ambient = Color.FromArgb(0x20, 0x20, 0x20);
			device.RenderState.SpecularEnable = true; //zapnuti spekularnich odlesku

			device.RenderState.FogColor = Color.Black;
			device.RenderState.FogStart = camNear;
			device.RenderState.FogEnd = camFar;
			device.RenderState.FogDensity = 0.05f;
			device.RenderState.FogVertexMode = FogMode.Exp;
			device.RenderState.FogEnable = true;

			//nastaveni nejlepsiho filtrovani textury
			setBestTextureFilters();

			//nastaveni projekcni matice
			device.Transform.Projection = Matrix.PerspectiveFovLH(camFov, (float) ClientSize.Width / ClientSize.Height, camNear, camFar);

			//vytvorime znova vsechny objekty, cimz se se znova inicializuji vertexy
			svet.ReInitialize(device, dat);

			//znovu nastavime svetla
			CreateLights();
		}

		public void CreateLights()
		{
			//nastaveni svitu z baterky
			device.Lights[0].Enabled = true;
			device.Lights[0].Type = LightType.Spot; //nastaveni typu svetla na spot
			device.Lights[0].Diffuse = Color.White; //nastaveni difuzni barvy svetla
			device.Lights[0].Specular = Color.Silver; //nastaveni spekularni barvy svetla
			device.Lights[0].Ambient = Color.FromArgb(0x40, 0x40, 0x40);
			device.Lights[0].Attenuation0 = 0.00025f; //0.00025f, 0.001f
			device.Lights[0].Attenuation2 = 0.009f; //0.009f, 0.01f
			device.Lights[0].InnerConeAngle = 32 * (float) Math.PI / 180; //vrcholovy uhel vnitrniho kuzele
			device.Lights[0].OuterConeAngle = 36 * (float) Math.PI / 180; //vrcholovy uhel vnejsiho kuzele
			device.Lights[0].Falloff = 1; //druh opadu svetla mezi kuzely
			device.Lights[0].Range = camFar; //dosah svetla
			device.Lights[0].Update();
		}

		//detekce nejlepsich filtru pro textury
		public void setBestTextureFilters()
		{
			// magfilter
			if (device.DeviceCaps.TextureFilterCaps.SupportsMagnifyGaussianQuad)
			{
				device.SamplerState[0].MagFilter = TextureFilter.GaussianQuad;
				Console.WriteLine("Using GaussianQuad MagFilter");
			}
			else if (device.DeviceCaps.TextureFilterCaps.SupportsMagnifyPyramidalQuad)
			{
				device.SamplerState[0].MagFilter = TextureFilter.PyramidalQuad;
				Console.WriteLine("Using PyramidalQuad MagFilter");
			}
			else if (device.DeviceCaps.TextureFilterCaps.SupportsMagnifyAnisotropic)
			{
				device.SamplerState[0].MagFilter = TextureFilter.Anisotropic;
				Console.WriteLine("Using Anisotropic MagFilter");
			}
			else if (device.DeviceCaps.TextureFilterCaps.SupportsMagnifyLinear)
			{
				device.SamplerState[0].MagFilter = TextureFilter.Linear;
				Console.WriteLine("Using Linear MagFilter");
			}
			else
			{
				device.SamplerState[0].MagFilter = TextureFilter.None;
				Console.WriteLine("Using None MagFilter");
			}

			// minfilter
			if (device.DeviceCaps.TextureFilterCaps.SupportsMinifyGaussianQuad)
			{
				device.SamplerState[0].MinFilter = TextureFilter.GaussianQuad;
				Console.WriteLine("Using GaussianQuad MinFilter");
			}
			else if (device.DeviceCaps.TextureFilterCaps.SupportsMinifyPyramidalQuad)
			{
				device.SamplerState[0].MinFilter = TextureFilter.PyramidalQuad;
				Console.WriteLine("Using PyramidalQuad MinFilter");
			}
			else if (device.DeviceCaps.TextureFilterCaps.SupportsMinifyAnisotropic)
			{
				device.SamplerState[0].MinFilter = TextureFilter.Anisotropic;
				Console.WriteLine("Using Anisotropic MinFilter");
			}
			else if (device.DeviceCaps.TextureFilterCaps.SupportsMinifyLinear)
			{
				device.SamplerState[0].MinFilter = TextureFilter.Linear;
				Console.WriteLine("Using Linear MinFilter");
			}
			else if (device.DeviceCaps.TextureFilterCaps.SupportsMinifyPoint)
			{
				device.SamplerState[0].MinFilter = TextureFilter.Point;
				Console.WriteLine("Using Point MinFilter");
			}
			else
			{
				device.SamplerState[0].MinFilter = TextureFilter.None;
				Console.WriteLine("Using None MinFilter");
			}

			// mipfilter
			if (device.DeviceCaps.TextureFilterCaps.SupportsMipMapLinear)
			{
				device.SamplerState[0].MipFilter = TextureFilter.Linear;
				Console.WriteLine("Using Linear MipFilter");
			}
			else
			{
				device.SamplerState[0].MipFilter = TextureFilter.Point;
				Console.WriteLine("Using Point MipFilter");
			}
		}

		/// <summary>
		/// Vycisti okno a vykresli scenu
		/// </summary>
		private void Render()
		{
			try
			{
				//if (!Focused && !drzKrysu) Thread.Sleep(50);
				// nejdriv provedeme vypocet fps
				// spustime tick, pokud neni
				if (tick == 0)
					tick = Environment.TickCount;

				dobaVykreslovani = ((float) (Environment.TickCount - LastTick)) / 1000;
				LastTick = Environment.TickCount;

				// vyprsela 1 vterina?
				float delay = Math.Abs((Environment.TickCount - tick) / 1000);
				if (delay > 0.5f)
				{
					fps = (int) (loopsCount / delay);
					loopsCount = 0;
					tick = Environment.TickCount;
				}

				//nacteni z klavesnice a mysi
				UpdateKeyboard();
				UpdateMouse();

				device.Clear(ClearFlags.Target | ClearFlags.ZBuffer, BackColor, 1, 0); /// vycisti pamet hloubky a barev
				device.BeginScene(); /// zacatek sceny

				//upravime pozice a smer svitu baterky
				device.Lights[0].Position = new Vector3(-svet.getHrac().x - 4 * (float) Math.Sin(-azimuth), 2.05f - 4 * (float) Math.Sin(zenith), -svet.getHrac().y - 4 * (float) Math.Cos(-azimuth));
				device.Lights[0].Direction = new Vector3((float) Math.Sin(-azimuth), (float) Math.Sin(zenith - 2 * Math.PI / 180), (float) Math.Cos(-azimuth));
				device.Lights[0].Update();

				//nastavime matici view
				device.Transform.View = Matrix.Identity;
				device.Transform.View *= Matrix.Translation(svet.getHrac().x, -1.85f, svet.getHrac().y);
				device.Transform.View *= Matrix.RotationY(azimuth);
				device.Transform.View *= Matrix.RotationX(zenith);

				//nastavime material vsech ploch
				device.Material = material;

				// prerender transformace view
				svet.preRender();

				svet.renderTime = dobaVykreslovani;
				//vykreslime svet

				if (bRenderWireFrame)
				{
					// kresleni drateneho modelu
					svet.noTexture = true;
					device.SetTexture(0, null);
					device.RenderState.Lighting = false;
					device.RenderState.FogEnable = false;

					device.RenderState.ColorWriteEnable = 0;
					device.RenderState.DepthBias = 0.0015f;
					svet.Render();
					device.RenderState.ColorWriteEnable = ColorWriteEnable.RedGreenBlueAlpha;
					device.RenderState.FillMode = FillMode.WireFrame;
					device.RenderState.DepthBias = 0.001f;
					svet.renderTime = 0;
					svet.Render();
					device.RenderState.FillMode = FillMode.Solid;
				}
				else
				{
					// kresleni solid modelu
					device.RenderState.Lighting = !bRenderNoLights;
					device.RenderState.FogEnable = !bRenderNoFog;
					svet.noTexture = bRenderNoTextures;
					if (bRenderNoTextures) device.SetTexture(0, null);
					svet.Render();
				}

				// postrender kresleni gui
				svet.postRender();
				//zobrazime fps
				printFPS(fps);

				device.EndScene(); //konec sceny
				device.Present(); //vykresleni 
				loopsCount++;
			}
			catch(Exception e)
			{
				Console.WriteLine(e);
			}
		}

		public bool InputInit()
		{
			//create keyboard device.
			keyboard = new Device(SystemGuid.Keyboard);
			if (keyboard == null)
			{
				Console.WriteLine("No keyboard found.");
				return false;
			}

			//create mouse device.
			mouse = new Device(SystemGuid.Mouse);
			if (mouse == null)
			{
				Console.WriteLine("No mouse found.");
				return false;
			}

			//Set mouse axis mode absolute.
			mouse.Properties.AxisModeAbsolute = true;

			//set cooperative level.
			keyboard.SetCooperativeLevel(this, CooperativeLevelFlags.NonExclusive | CooperativeLevelFlags.Background);

			mouse.SetCooperativeLevel(this, CooperativeLevelFlags.NonExclusive | CooperativeLevelFlags.Background);

			mouse.SetDataFormat(DeviceDataFormat.Mouse);

			//Acquire devices for capturing.
			keyboard.Acquire();
			mouse.Acquire();

			return true;
		}

		protected override void OnKeyPress(KeyPressEventArgs e)
		{
			char key = Char.ToLower(e.KeyChar);

			if (key == 'e' || key == (char) Keys.Enter)
			{
				//otevreni dveri, pouziti objectu
				svet.UseObject((int) ((-svet.getHrac().x / 2) - (float) Math.Sin(azimuth)), (int) ((-svet.getHrac().y / 2) + (float) Math.Cos(azimuth)));
			}
			else if (key == 'h' || key == 'r')
			{
				//nastvime hrace na startovni pozici
				svet.setHracHP();
				azimuth = 0;
				zenith = 0;
				mouseX = 0;
				mouseY = 10;
			}
			else if (key == 'x')
			{
				//ukonceni enginu
				Close();
				Application.Exit();
			}
			else if (key == 'k')
			{
				//drzet / nedrzet krysu (mys)
				MouseState state = mouse.CurrentMouseState;
				drzKrysu = !drzKrysu;
				if (drzKrysu)
					Cursor.Hide();
				else
					Cursor.Show();
			}
			else if (key == 'u')
			{
				//invertace mysi
				invertMouse = !invertMouse;
			}
			else if (key == 'l')
			{
				//nacteni nove mapy
				try
				{
					if (drzKrysu)
						Cursor.Show();
					drzKrysu = false;

					OpenFileDialog dialog = new OpenFileDialog();
					dialog.Title = "Load map";
					dialog.Filter = "Map file (*.map, *.txt)|*.map;*.txt";
					dialog.FilterIndex = 0;
					if (dialog.ShowDialog() == DialogResult.OK)
					{
						svet = new Svet(device, dat);
						svet.LoadMap(dialog.FileName);
						svet.SetRenderForm(this);

						kolize = new Kolize(svet);
						azimuth = 0;
						zenith = 0;
						mouseX = 0;
						mouseY = 10;
					}

					drzKrysu = true;
					Cursor.Hide();
				}
				catch(Exception ex)
				{
					Console.WriteLine("Nastala chyba s dialogem...");
					Console.WriteLine(ex);
				}
			}
			else if (key == '1' || key == '+')
			{
				bRenderWireFrame = !bRenderWireFrame;
			}
			else if (key == '2' || key == '')
			{
				bRenderNoLights = !bRenderNoLights;
			}
			else if (key == '3' || key == '')
			{
				bRenderNoFog = !bRenderNoFog;
			}
			else if (key == '4' || key == '')
			{
				bRenderNoTextures = !bRenderNoTextures;
			}
			else if (key == (char) Keys.Tab)
			{
				svet.ToggleHelp();
			}
		}

		private void UpdateKeyboard()
		{
			int kombinace_klaves = 0;
			Hrac hrac = svet.getHrac();
			Hrac hrac_novy = hrac;
			float rychlostPohybu = 2.5f * dobaVykreslovani;

			Key[] keys = keyboard.GetPressedKeys();

			if (!drzKrysu || !svet.isPlayerMovable)
				return;

			//Capture pressed keys.
			foreach(Key k in keys)
			{
				if (k == Key.W || k == Key.Up)
				{
					hrac_novy.y -= rychlostPohybu * (float) Math.Cos(azimuth);
					hrac_novy.x += rychlostPohybu * (float) Math.Sin(azimuth);
					kombinace_klaves |= 1;
				}
				else if (k == Key.S || k == Key.Down)
				{
					hrac_novy.y += rychlostPohybu * (float) Math.Cos(azimuth);
					hrac_novy.x -= rychlostPohybu * (float) Math.Sin(azimuth);
					kombinace_klaves |= 1;
				}
				else if (k == Key.A || k == Key.Left)
				{
					hrac_novy.y += rychlostPohybu * (float) (Math.Cos(azimuth - Math.PI / 2));
					hrac_novy.x -= rychlostPohybu * (float) (Math.Sin(azimuth - Math.PI / 2));
					kombinace_klaves |= 2;
				}
				else if (k == Key.D || k == Key.Right)
				{
					hrac_novy.y += rychlostPohybu * (float) (Math.Cos(azimuth + Math.PI / 2));
					hrac_novy.x -= rychlostPohybu * (float) (Math.Sin(azimuth + Math.PI / 2));
					kombinace_klaves |= 2;
				}
			}
			if (kombinace_klaves == 0)
				return;

			if ((kombinace_klaves & 1) == 1 && (kombinace_klaves & 2) == 2)
			{
				hrac_novy.x = hrac.x + (hrac_novy.x - hrac.x) / (float) Math.Sqrt(2);
				hrac_novy.y = hrac.y + (hrac_novy.y - hrac.y) / (float) Math.Sqrt(2);
			}
			hrac = kolize.UpravPohyb(hrac, hrac_novy);

			svet.setHrac(hrac);
		}

		private void UpdateMouse()
		{
			if (!drzKrysu || !svet.isPlayerMovable)
				return;

			//Get Mouse State.
			MouseState state = mouse.CurrentMouseState;
			if (state.X == 0 && state.Y == 0)
				return;

			if (state.X != 0)
			{
				mouseX += state.X;
				azimuth = (float) (((float) -mouseX / Width) * (2 * Math.PI)); /// vypocet azimutu a zenitu na zaklade abs. polohy kurzoru	
			}
			if (state.Y != 0)
			{
				mouseY += state.Y;
				zenith = (float) (((float) -mouseY / (Height)) * Math.PI / 2);
				if (invertMouse)
					zenith = -zenith;

				if (zenith > Math.PI / 2)
				{
					zenith = (float) Math.PI / 2;
					mouseY -= state.Y;
				}
				if (zenith < -Math.PI / 2)
				{
					zenith = (float) -Math.PI / 2;
					mouseY -= state.Y;
				}
			}
			if (drzKrysu && (state.X != 0 || state.Y != 0))
				Cursor.Position = new Point(Left + Width / 2, Top + Height / 2);
		}

		private void printFPS(int _fps)
		{
			d3dFont.DrawText(null, //Advanced parameter
			                 "FPS: " + _fps /*+ "; x: " + (-hrac.x) + "; y:" + (-hrac.y) + "; map: " + svet.map[(int) (-hrac.x)/2][(int) (-hrac.y)/2] + "; lx: " + (float) Math.Cos(-azimuth) + "; ly: " + (float) Math.Sin(zenith) + "; lz: " + (float) Math.Sin(-azimuth)*/, // Text to render
			                 ClientRectangle, //Clip text to this rectangle
			                 DrawTextFormat.Left | //Align text to the left of the window
			                 DrawTextFormat.Top | //and to the top
			                 DrawTextFormat.WordBreak, //And break lines if necessary
			                 Color.GreenYellow); //What color to draw the text in
		}

		public void drawText(Sprite sprite, string text, Rectangle rect, DrawTextFormat format, Color color)
		{
			d3dFont.DrawText(sprite, text, rect, format, color);
		}

		protected override void WndProc(ref Message m)
		{
			if (m.Msg == 0x000F)
			{
				Render();
				SendNotifyMessage(Handle, 0x000F, IntPtr.Zero, IntPtr.Zero);
			}
			else
				base.WndProc(ref m);
		}

		/// <summary>
		/// Po spusteni se pokusi inicializovat Direct3D. V pripade uspechu
		/// Prekresluje okno a zpracovava zpravy, dokud existuje okno.
		/// </summary>
		[STAThreadAttribute]
		public static void Main()
		{
			RenderForm frm = new RenderForm(); //vytvoreni formulare
			if (frm.D3DInit() && frm.InputInit()) //pokud se podari inicializovat d3d
			{
				frm.Show(); //zobrazeni formu
				Application.Run();
			}
			if (device != null)
				device.Dispose();
			frm = null;
		}
	}
}